__version__ = "2.2.4"

from .api import AIAPI, OpenAIAPI
from libbs.api import DecompilerInterface


def create_plugin(*args, **kwargs):

    #
    # OpenAI API (ChatGPT)
    #

    openai_api = OpenAIAPI(delay_init=True)
    # create context menus for prompts
    gui_ctx_menu_actions = {
        f"DAILA/OpenAI/{prompt_name}": (prompt.desc, getattr(openai_api, prompt_name))
        for prompt_name, prompt in openai_api.prompts_by_name.items()
    }
    # create context menus for others
    gui_ctx_menu_actions["DAILA/OpenAI/update_api_key"] = ("Update API Key", openai_api.ask_api_key)

    #
    # VarModel API (local variable renaming)
    #

    from varbert.api import VariableRenamingAPI
    var_api = VariableRenamingAPI(delay_init=True)
    # add single interface, which is to rename variables
    gui_ctx_menu_actions["DAILA/VarBERT/varbert_rename_vars"] = ("Suggest new variable names", var_api.query_model)

    #
    # Decompiler Plugin Registration
    #

    force_decompiler = kwargs.pop("force_decompiler", None)
    deci = DecompilerInterface.discover(
        force_decompiler=force_decompiler,
        # decompiler-creation args
        plugin_name="DAILA",
        init_plugin=True,
        gui_ctx_menu_actions=gui_ctx_menu_actions,
        ui_init_args=args,
        ui_init_kwargs=kwargs
    )

    openai_api.init_decompiler_interface(decompiler_interface=deci)
    if var_api is not None:
        var_api.init_decompiler_interface(decompiler_interface=deci)

    return deci.gui_plugin



# Run the DAILA server and open the DAILA selector UI.
# @author mahaloz
# @category AI
# @menupath Tools.DAILA.Start DAILA Backend


# replace this with the command to run your plugin remotely
library_command = "dailalib -s ghidra"
# replace this imporate with an import of your plugin's create_plugin function
def create_plugin(*args, **kwargs):
    from dailalib import create_plugin as _create_plugin
    return _create_plugin(*args, **kwargs)

# =============================================================================
# LibBS generic plugin loader (don't touch)
# =============================================================================

import sys
# Python 2 has special requirements for Ghidra, which forces us to use a different entry point
# and scope for defining plugin entry points
if sys.version[0] == "2":
    # Do Ghidra Py2 entry point
    import subprocess
    from libbs_vendored.ghidra_bridge_server import GhidraBridgeServer
    full_command = "python3 -m " + library_command

    GhidraBridgeServer.run_server(background=True)

    failed_py3 = False
    try:
        process = subprocess.Popen(full_command.split(" "))
    except:
        failed_py3 = True

    if failed_py3:
        print("Couldn't find python3 in your path. Trying python...")
        full_command = full_command.replace("python3", "python")
        process = subprocess.Popen(full_command.split(" "))

    if process.poll() is not None:
        raise RuntimeError(
            "Failed to run the Python3 backed. It's likely Python3 is not in your Path inside Ghidra.")
else:
    # Try plugin discovery for other decompilers
    try:
        import idaapi
        has_ida = True
    except ImportError:
        has_ida = False
    try:
        import angrmanagement
        has_angr = True
    except ImportError:
        has_angr = False

    if not has_ida and not has_angr:
        create_plugin()
    elif has_angr:
        from angrmanagement.plugins import BasePlugin
        class AngrBSPluginThunk(BasePlugin):
            def __init__(self, workspace):
                super().__init__(workspace)
                globals()["workspace"] = workspace
                self.plugin = create_plugin()

            def teardown(self):
                pass


def PLUGIN_ENTRY(*args, **kwargs):
    """
    This is the entry point for IDA to load the plugin.
    """
    return create_plugin(*args, **kwargs)



import typing
from typing import Optional
import os

from openai import OpenAI
import tiktoken

from dailalib.api.ai_api import AIAPI
if typing.TYPE_CHECKING:
    from dailalib.api.openai.prompt import Prompt


class OpenAIAPI(AIAPI):
    prompts_by_name = []
    DEFAULT_MODEL = "gpt-4"
    MODEL_TO_TOKENS = {
        "gpt-4": 8000,
        "gpt-3.5-turbo": 4096
    }

    # replacement strings for API calls
    def __init__(self, api_key: Optional[str] = None, model: str = DEFAULT_MODEL, prompts: Optional[list] = None, **kwargs):
        super().__init__(**kwargs)
        self._api_key = None
        self._openai_client: OpenAI = None
        self.api_key = api_key or os.getenv("OPENAI_API_KEY")
        self.model = model

        # delay prompt import
        from dailalib.api.openai.prompts import PROMPTS
        prompts = prompts + PROMPTS if prompts else PROMPTS
        self.prompts_by_name = {p.name: p for p in prompts}

    def __dir__(self):
        return list(super().__dir__()) + list(self.prompts_by_name.keys())

    def __getattribute__(self, item):
        if item in object.__getattribute__(self, "prompts_by_name"):
            prompt_obj: "Prompt" = self.prompts_by_name[item]
            prompt_obj.ai_api = self
            return prompt_obj.query_model
        else:
            return object.__getattribute__(self, item)

    @property
    def api_key(self):
        return self._api_key

    @api_key.setter
    def api_key(self, value):
        self._api_key = value
        if self._api_key:
            self._openai_client = OpenAI(api_key=self._api_key)

    def ask_api_key(self, *args, **kwargs):
        self.api_key = self._dec_interface.gui_ask_for_string("Enter you OpenAI API Key:", title="DAILA")

    def query_model(
        self,
        prompt: str,
        model: Optional[str] = None,
        max_tokens=None,
    ):
        if not self._openai_client:
            raise ValueError("You must provide an API key before querying the model.")

        response = self._openai_client.chat.completions.create(
            model=model or self.model,
            messages=[
                {"role": "user", "content": prompt}
            ],
            max_tokens=max_tokens,
            timeout=60,
        )

        try:
            answer = response.choices[0].message.content
        except (KeyError, IndexError) as e:
            answer = None

        return answer

    @staticmethod
    def estimate_token_amount(content: str, model=DEFAULT_MODEL):
        enc = tiktoken.encoding_for_model(model)
        tokens = enc.encode(content)
        return len(tokens)

    @staticmethod
    def content_fits_tokens(content: str, model=DEFAULT_MODEL):
        max_token_count = OpenAIAPI.MODEL_TO_TOKENS[model]
        token_count = OpenAIAPI.estimate_token_amount(content, model=model)
        return token_count <= max_token_count - 1000

    @staticmethod
    def fit_decompilation_to_token_max(decompilation: str, delta_step=10, model=DEFAULT_MODEL):
        if OpenAIAPI.content_fits_tokens(decompilation, model=model):
            return decompilation

        dec_lines = decompilation.split("\n")
        last_idx = len(dec_lines) - 1
        # should be: [func_prototype] + [nop] + [mid] + [nop] + [end_of_code]
        dec_lines = dec_lines[0:2] + ["// ..."] + dec_lines[delta_step:last_idx-delta_step] + ["// ..."] + dec_lines[-2:-1]
        decompilation = "\n".join(dec_lines)

        return OpenAIAPI.fit_decompilation_to_token_max(decompilation, delta_step=delta_step, model=model)